<?php
/**
 * This file is part of FacturaScripts
 * Copyright (C) 2013-2022 Carlos Garcia Gomez <carlos@facturascripts.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace FacturaScripts\Core\Model;

use FacturaScripts\Dinamic\Model\Empresa as DinEmpresa;
use FacturaScripts\Dinamic\Model\Page as DinPage;

/**
 * Usuario de FacturaScripts.
 *
 * @author Carlos García Gómez <carlos@facturascripts.com>
 */
class User extends Base\ModelClass
{

    use Base\ModelTrait;
    use Base\CompanyRelationTrait;
    use Base\PasswordTrait;

    const DEFAULT_LEVEL = 2;

    /**
     * true -> user is admin.
     *
     * @var bool
     */
    public $admin;

    /**
     * @var string
     */
    public $codagente;

    /**
     * @var string
     */
    public $codalmacen;

    /**
     * @var string
     */
    public $creationdate;

    /**
     * user's email.
     *
     * @var string
     */
    public $email;

    /**
     * true -> user enabled.
     *
     * @var bool
     */
    public $enabled;

    /**
     * Homepage.
     *
     * @var string
     */
    public $homepage;

    /**
     * Language code.
     *
     * @var string
     */
    public $langcode;

    /**
     * Last activity date.
     *
     * @var string
     */
    public $lastactivity;

    /**
     * Last IP used.
     *
     * @var string
     */
    public $lastip;

    /**
     * Indicates the level of security that the user can access.
     *
     * @var integer
     */
    public $level;

    /**
     * Session key, saved also in cookie. Regenerated when user log in.
     *
     * @var string
     */
    public $logkey;

    /**
     * Primary key. Varchar (50).
     *
     * @var string
     */
    public $nick;

    public function clear()
    {
        parent::clear();
        $this->codalmacen = $this->toolBox()->appSettings()->get('default', 'codalmacen');
        $this->creationdate = date(self::DATE_STYLE);
        $this->enabled = true;
        $this->idempresa = $this->toolBox()->appSettings()->get('default', 'idempresa', 1);
        $this->langcode = FS_LANG;
        $this->level = self::DEFAULT_LEVEL;
    }

    public function delete(): bool
    {
        if ($this->count() === 1) {
            // prevent delete all users
            $this->toolBox()->i18nLog()->error('cant-delete-last-user');
            return false;
        }

        return parent::delete();
    }

    public function install(): string
    {
        // we need this models to be checked before
        new DinPage();
        new DinEmpresa();

        $nick = defined('FS_INITIAL_USER') ? FS_INITIAL_USER : 'admin';
        $pass = defined('FS_INITIAL_PASS') ? FS_INITIAL_PASS : 'admin';
        $email = filter_var($this->nick, FILTER_VALIDATE_EMAIL) ? $this->nick : '';
        $this->toolBox()->i18nLog()->notice('created-default-admin-account', ['%nick%' => $nick, '%pass%' => $pass]);
        return 'INSERT INTO ' . static::tableName() . ' (nick,password,email,admin,enabled,idempresa,codalmacen,langcode,homepage,level)'
            . " VALUES ('" . $nick . "','" . password_hash($pass, PASSWORD_DEFAULT) . "','" . $email
            . "',TRUE,TRUE,'1','1','" . FS_LANG . "','Wizard','99');";
    }

    /**
     * Generates a new login key for the user. It also updates lastactivity
     * and last IP.
     *
     * @param string $ipAddress
     *
     * @return string
     */
    public function newLogkey(string $ipAddress): string
    {
        $this->updateActivity($ipAddress);
        $this->logkey = $this->toolBox()->utils()->randomString(99);
        return $this->logkey;
    }

    public static function primaryColumn(): string
    {
        return 'nick';
    }

    public static function tableName(): string
    {
        return 'users';
    }

    public function test(): bool
    {
        $this->nick = trim($this->nick);
        if (1 !== preg_match("/^[A-Z0-9_@\+\.\-]{3,50}$/i", $this->nick)) {
            $this->toolBox()->i18nLog()->error(
                'invalid-alphanumeric-code',
                ['%value%' => $this->nick, '%column%' => 'nick', '%min%' => '3', '%max%' => '50']
            );
            return false;
        }

        $this->email = $this->toolBox()->utils()->noHtml(mb_strtolower($this->email, 'UTF8'));
        if ($this->email && false === filter_var($this->email, FILTER_VALIDATE_EMAIL)) {
            $this->toolBox()->i18nLog()->warning('not-valid-email', ['%email%' => $this->email]);
            $this->email = null;
            return false;
        }

        if (empty($this->creationdate)) {
            $this->creationdate = date(self::DATE_STYLE);
        }

        if (empty($this->lastactivity)) {
            $this->lastactivity = null;
        }

        if ($this->admin) {
            $this->level = 99;
        } elseif ($this->level === null) {
            $this->level = 0;
        }

        return $this->testPassword() && $this->testAgent() && $this->testWarehouse() && parent::test();
    }

    /**
     * Updates last ip address and last activity property.
     *
     * @param string $ipAddress
     */
    public function updateActivity(string $ipAddress)
    {
        $this->lastactivity = date(self::DATETIME_STYLE);
        $this->lastip = $ipAddress;
    }

    /**
     * Verifies the login key.
     *
     * @param string $value
     *
     * @return bool
     */
    public function verifyLogkey(string $value): bool
    {
        return $this->logkey === $value;
    }

    protected function saveInsert(array $values = []): bool
    {
        $result = parent::saveInsert($values);
        if ($result && false === $this->admin) {
            $this->setNewRole();
        }

        return $result;
    }

    /**
     * Assigns the default role to this user.
     */
    protected function setNewRole()
    {
        $role = new Role();
        $code = $this->toolBox()->appSettings()->get('default', 'codrole');
        if (false === $role->loadFromCode($code)) {
            return;
        }

        $roleUser = new RoleUser();
        $roleUser->codrole = $role->codrole;
        $roleUser->nick = $this->nick;
        $roleUser->save();

        // set user homepage
        foreach ($roleUser->getRoleAccess() as $roleAccess) {
            $this->homepage = $roleAccess->pagename;
            if ('List' == substr($this->homepage, 0, 4)) {
                break;
            }
        }
        $this->save();
    }

    protected function testAgent(): bool
    {
        if (empty($this->codagente)) {
            $this->codagente = null;
            return true;
        }

        $agent = new Agente();
        if (false === $agent->loadFromCode($this->codagente)) {
            $this->codagente = null;
        }

        return true;
    }

    protected function testWarehouse(): bool
    {
        $appSettings = $this->toolBox()->appSettings();

        if (empty($this->codalmacen)) {
            $this->codalmacen = $appSettings->get('default', 'codalmacen');
            $this->idempresa = $appSettings->get('default', 'idempresa');
            return true;
        }

        $warehouse = new Almacen();
        if (false === $warehouse->loadFromCode($this->codalmacen) || $warehouse->idempresa != $this->idempresa) {
            $this->codalmacen = $appSettings->get('default', 'codalmacen');
            $this->idempresa = $appSettings->get('default', 'idempresa');
        }

        return true;
    }
}
